/* Anonymous Read Pipe Example - by Mike Caetano (prionx@juno.com)

A GUI app lacks a console, therefore no standard io.
In order to read the output of console programs launched as child processes
to a GUI process, the standard io handles have to be redirected to a pipe.

The availability of pipes is limited under Windows 9x,
so an anonymous pipe must be used in order to read the console output
on both Win9x and WinNT.

Once the anonymous pipe has been created, the handle has to be duplicated
so that the child process can inherit it as it's own and write to it.
The GUI parent can then read from the other end of the pipe and transfer
the data to an appropriate GUI control, in this case an edit control.

// ---------------------------------------------------------------------------
Standard disclaimer:

I've done my best to insure the code works and is relatively bug free,
but I accept no responsibilities for any problems it might cause to your machine.
Use at your own risk. Etc...

*/
#include <windows.h>
#include "resource.h"

#define BUFSIZE 0x10000 // aka 64kb - max edit box capacity

// ---------------------------------------------------------------------------
// prototypes

LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);
VOID ReportError(DWORD errno);

// ---------------------------------------------------------------------------
// global vars

TCHAR szAppName[] = TEXT ("Anonymous Read Pipe Example");


// ---------------------------------------------------------------------------
// generic error report function
VOID ReportError(DWORD errno)
{
	LPVOID lpMsgBuf;

	DWORD msglen = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM
		    , NULL
		    , errno
		    , MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT)
		    , (LPTSTR) &lpMsgBuf
		    , 0, NULL);

	if ( msglen > 0 ) {
		MessageBox(GetActiveWindow(), lpMsgBuf, "Error Report", MB_OK | MB_ICONEXCLAMATION);
	}

	LocalFree( lpMsgBuf );

}

// ---------------------------------------------------------------------------
//
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{

	// these have to remain consistent from call to call
	static HANDLE hInputRead = INVALID_HANDLE_VALUE;
	static HANDLE hInputWrite = INVALID_HANDLE_VALUE;
	static HANDLE hDupInputWrite = INVALID_HANDLE_VALUE;
    static BOOL bPipeStatus = FALSE;
	static STARTUPINFO si;
	static HWND hwndEdit;

	// these don't have to remain consistent and shouldn't
	PROCESS_INFORMATION pi;
	TCHAR szInput[BUFSIZE];
	DWORD dwExitCode = 0;
	DWORD dwRead = 0;

	switch (message)
	{

	case WM_CREATE :

		hwndEdit = CreateWindow( TEXT("edit"), NULL,
			WS_CHILD | WS_VISIBLE| WS_BORDER | WS_VSCROLL |
			    ES_LEFT | ES_MULTILINE |  ES_AUTOHSCROLL | ES_AUTOVSCROLL,
			0, 0,
			0, 0,
			hwnd,
			(HMENU) IDC_EDIT,
			((LPCREATESTRUCT) lParam)->hInstance, NULL);

		if ( hwndEdit == NULL ) {
			ReportError(GetLastError());
			break;
		}

		// explicitly limit the capacity of the edit control
		SendMessage (hwndEdit, EM_LIMITTEXT, BUFSIZE, 0L);

		//  create an anonymous pipe for the child process to inherit
		bPipeStatus = CreatePipe(&hInputRead, &hInputWrite, NULL, 0 );
		if( !bPipeStatus ) {
			ReportError(GetLastError());
			break;
		}

		//  make the write end inheritable
		bPipeStatus = DuplicateHandle( GetCurrentProcess()
						, hInputWrite
						, GetCurrentProcess()
						, &hDupInputWrite
						, 0
						, TRUE
						, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS );

		if( !bPipeStatus ) {
			ReportError(GetLastError());
			break;
		}

		// set up the startup info for the various child process
		ZeroMemory(&si, sizeof(STARTUPINFO));
		si.cb = sizeof(STARTUPINFO);
		si.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
		si.wShowWindow = SW_HIDE;		// hide child process
		si.hStdOutput = hDupInputWrite; // assign inheritable pipe end
		si.hStdError = hDupInputWrite;  // redirect error messages too

		// default to the standard input handle
		si.hStdInput = GetStdHandle(STD_INPUT_HANDLE);

		return 0;

	case WM_SETFOCUS:
		SetFocus(hwndEdit);
		return 0;

	case WM_SIZE:
		MoveWindow(hwndEdit, 0, 0, LOWORD (lParam), HIWORD (lParam), TRUE) ;
		return 0;

	case WM_CTLCOLOREDIT:
		// black background, green text
		SetBkColor((HDC)wParam, RGB(0,0,0));
		SetTextColor((HDC)wParam, RGB(0,204,0));
		// return stock object to let the system deal with clean up
		return (LRESULT)GetStockObject(BLACK_BRUSH);

	case WM_COMMAND:

		switch (LOWORD (wParam))
		{
		case IDM_LCC:

			// set the transitory variables to a known state
			ZeroMemory(&pi, sizeof(PROCESS_INFORMATION) );
			ZeroMemory(szInput, BUFSIZE);
			dwExitCode = 0;
			dwRead = 0;

			// this will fail if lcc isn't in the path
			if ( !CreateProcess( NULL
						, "lcc"
						, NULL
						, NULL
						, TRUE
						, 0
						, NULL
						, NULL
						, &si
						, &pi ) )
			{
				ReportError(GetLastError());
				break;
			}

			// wait for the process to complete
			WaitForSingleObject(pi.hProcess, INFINITE);

			// check for hang up errors
			if ( !GetExitCodeProcess(pi.hProcess, &dwExitCode) ) {
				ReportError(GetLastError());
				break;
			}

			// close these handles
			CloseHandle(pi.hThread);
			CloseHandle(pi.hProcess);

			// read output into a buffer
			ReadFile(hInputRead, szInput, BUFSIZE, &dwRead, NULL);

			// if there are characters in the buffer send them to the edit control
			if ( dwRead > 0 ) {
				SendMessage(hwndEdit, WM_SETTEXT, 0, (LPARAM)szInput);
			}

			break;

		case IDM_LRC:

			// set the transitory variables to a known state
			ZeroMemory(&pi, sizeof(PROCESS_INFORMATION) );
			ZeroMemory(szInput, BUFSIZE);
			dwExitCode = 0;
			dwRead = 0;

			// this will fail if lrc isn't in the path
			if ( !CreateProcess( NULL
						, "lrc"
						, NULL
						, NULL
						, TRUE
						, 0
						, NULL
						, NULL
						, &si
						, &pi ) )
			{
				ReportError(GetLastError());
				break;
			}

			// wait for the process to complete
			WaitForSingleObject(pi.hProcess, INFINITE);

			// check for hang up errors
			if ( !GetExitCodeProcess(pi.hProcess, &dwExitCode) ) {
				ReportError(GetLastError());
				break;
			}

			// close these handles
			CloseHandle(pi.hThread);
			CloseHandle(pi.hProcess);

			// read output into a buffer
			ReadFile(hInputRead, szInput, BUFSIZE, &dwRead, NULL);

			// if there are characters in the buffer send them to the edit control
			if ( dwRead > 0 ) {
				SendMessage(hwndEdit, WM_SETTEXT, 0, (LPARAM)szInput);
			}

			break;


		case IDM_PEDUMP:

			// set the transitory variables to a known state
			ZeroMemory(&pi, sizeof(PROCESS_INFORMATION) );
			ZeroMemory(szInput, BUFSIZE);
			dwExitCode = 0;
			dwRead = 0;

			// this will fail if pedump isn't in the path
			if ( !CreateProcess( NULL
						, "pedump"
						, NULL
						, NULL
						, TRUE
						, 0
						, NULL
						, NULL
						, &si
						, &pi ) )
			{
				ReportError(GetLastError());
				break;
			}

			// wait for the process to complete
			WaitForSingleObject(pi.hProcess, INFINITE);

			// check for hang up errors
			if ( !GetExitCodeProcess(pi.hProcess, &dwExitCode) ) {
				ReportError(GetLastError());
				break;
			}

			// close these handles
			CloseHandle(pi.hThread);
			CloseHandle(pi.hProcess);

			// read output into a buffer
			ReadFile(hInputRead, szInput, BUFSIZE, &dwRead, NULL);

			// if there are characters in the buffer send them to the edit control
			if ( dwRead > 0 ) {
				SendMessage(hwndEdit, WM_SETTEXT, 0, (LPARAM)szInput);
			}

			break;

		case IDM_EXIT:
			SendMessage(hwnd, WM_CLOSE, 0, 0);
			return 0;
		}
		break;

	case WM_CLOSE :

		// if there are pipes close them
		if ( bPipeStatus ) {
			CloseHandle(hDupInputWrite);
			CloseHandle(hInputWrite);
			CloseHandle(hInputRead);
		}

		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(hwnd, message, wParam, lParam);

}

// ---------------------------------------------------------------------------
//
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,PSTR szCmdLine, int iCmdShow)
{
	HWND     hwnd;
	MSG      msg;
	WNDCLASS wndclass;
	ZeroMemory(&wndclass, sizeof(wndclass));
	wndclass.style         = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc   = WndProc;
	wndclass.hInstance     = hInstance;
	wndclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
	wndclass.lpszMenuName  = MAKEINTRESOURCE(IDM_MAINMENU);
	wndclass.lpszClassName = szAppName;

	if ( !RegisterClass(&wndclass) )
	{
		ReportError(GetLastError());
		return 0;
	}

	hwnd = CreateWindow(szAppName, szAppName,
				WS_OVERLAPPEDWINDOW,
				CW_USEDEFAULT, CW_USEDEFAULT,
				CW_USEDEFAULT, CW_USEDEFAULT,
				NULL, NULL,	hInstance, NULL);

	if ( hwnd == NULL ) {
		ReportError(GetLastError());
		return 0;
	}

	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd) ;

	while ( GetMessage(&msg, NULL, 0, 0) )
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return msg.wParam;

}

